import logging
import os
import sys
import time
import datetime
import traceback
import momonga

from pprint import pprint
from momonga import EchonetPropertyCode as EPC

log_fmt = logging.Formatter('%(asctime)s | %(levelname)s | %(name)s - %(message)s')
log_hnd = logging.StreamHandler()
log_hnd.setFormatter(log_fmt)
momonga.logger.addHandler(log_hnd)
momonga.logger.setLevel(logging.DEBUG)
momonga.session_manager_logger.addHandler(log_hnd)
momonga.session_manager_logger.setLevel(logging.DEBUG)
momonga.sk_wrapper_logger.addHandler(log_hnd)
momonga.sk_wrapper_logger.setLevel(logging.DEBUG)

# set the following environment values before run.
rbid = os.environ.get('MOMONGA_ROUTEB_ID')
pwd = os.environ.get('MOMONGA_ROUTEB_PASSWORD')
dev = os.environ.get('MOMONGA_DEV_PATH')
baudrate = os.environ.get('MOMONGA_DEV_BAUDRATE')

exit_code = 1

if rbid is None:
    print('Set a Route-B ID.', file=sys.stderr)
    exit(exit_code)
elif pwd is None:
    print('Set a Route-B password.', file=sys.stderr)
    exit(exit_code)
elif dev is None:
    print('Set the path of a Wi-SUN device.', file=sys.stderr)
    exit(exit_code)

if baudrate is None:
    baudrate = 115200
else:
    baudrate = int(baudrate)

while True:
    try:
        with momonga.Momonga(rbid, pwd, dev, baudrate) as mo:
            print('---- operation status of smart meter ----')
            res = mo.get_operation_status()
            if res is True:
                print('turned on')
            elif res is False:
                print('turned off')
            else:
                print('unknown')
            print('----')
            time.sleep(5)

            print('---- installation location ----')
            res = mo.get_installation_location()
            print(res)
            print('----')
            time.sleep(5)

            print('---- standard version information ----')
            res = mo.get_standard_version()
            print(res)
            print('----')
            time.sleep(5)

            print('---- fault status ----')
            res = mo.get_fault_status()
            if res is True:
                print('fault occurred')
            elif res is False:
                print('no fault occurred')
            else:
                print('unknown')
            print('----')

            print('---- manufacturer code ----')
            res = mo.get_manufacturer_code()
            print(res)
            print('----')
            time.sleep(5)

            print('---- serial number ----')
            res = mo.get_serial_number()
            print(res)
            print('----')
            time.sleep(5)

            print('---- current time setting ----')
            res = mo.get_current_time_setting()
            print(res)
            print('----')
            time.sleep(5)

            print('---- current date setting ----')
            res = mo.get_current_date_setting()
            print(res)
            print('----')
            time.sleep(5)

            print('---- properties for status notification ----')
            res = mo.get_properties_for_status_notification()
            print(res)
            print('----')
            time.sleep(5)

            print('---- properties to set values ----')
            res = mo.get_properties_to_set_values()
            print(res)
            print('----')
            time.sleep(5)

            print('---- properties to get values ----')
            res = mo.get_properties_to_get_values()
            print(res)
            print('----')
            time.sleep(5)

            print('---- number of effective digits for cumulative energy ----')
            res = mo.get_number_of_effective_digits_for_cumulative_energy()
            print(res)
            print('----')
            time.sleep(5)

            print('---- measured cumulative energy (normal direction) [kWh] ----')
            res = mo.get_measured_cumulative_energy()
            print(res, 'kWh')
            print('----')
            time.sleep(5)

            print('---- measured cumulative energy (reverse direction) [kWh] ----')
            res = mo.get_measured_cumulative_energy(reverse=True)
            print(res, 'kWh')
            print('----')
            time.sleep(5)

            print('---- historical cumulative energy 1 (normal direction) [kWh] ----')
            res = mo.get_historical_cumulative_energy_1()
            pprint(res)
            print('----')
            time.sleep(5)

            print('---- historical cumulative energy 1 (reverse direction) [kWh] ----')
            res = mo.get_historical_cumulative_energy_1(reverse=True)
            pprint(res)
            print('----')
            time.sleep(5)

            print('---- instantaneous power [W] ----')
            res = mo.get_instantaneous_power()
            print(res, 'W')
            print('----')
            time.sleep(5)

            print('---- instantaneous current [A] ----')
            res = mo.get_instantaneous_current()
            print(res, 'A')
            print('----')
            time.sleep(5)

            print('---- cumulative energy measured at fixed time (normal direction) [kWh] ----')
            res = mo.get_cumulative_energy_measured_at_fixed_time()
            pprint(res)
            print('----')
            time.sleep(5)

            print('---- cumulative energy measured at fixed time (reverse direction) [kWh] ----')
            res = mo.get_cumulative_energy_measured_at_fixed_time(reverse=True)
            pprint(res)
            print('----')
            time.sleep(5)

            print('---- historical_cumulative_energy_2 [kWh] ----')
            res = mo.get_historical_cumulative_energy_2()
            pprint(res)
            print('----')
            time.sleep(5)

            print('---- setting parameters with request_to_set() ----')
            now = datetime.datetime.now()
            mo.request_to_set(
                day_for_historical_data_1={'day': 0},
                time_for_historical_data_2={'timestamp': now, 'num_of_data_points': 12},
            )
            print('----')
            time.sleep(5)

            print('---- testing all EchonetPropertyCode using request_to_get() one by one ----')
            all_codes = [e for e in momonga.EchonetPropertyCode]
            for epc_req in all_codes:
                try:
                    epc, r = mo.request_to_get({epc_req}).popitem()
                    print(f'epc: {epc.name}, result: {r}')
                except momonga.MomongaResponseNotPossible:
                    print(f'epc: {epc_req.name} not possible')
            print('----')
            time.sleep(5)

            print('---- requesting with 4 EPCs using request_to_get() at once ----')
            res = mo.request_to_get({
                EPC.instantaneous_power,
                EPC.instantaneous_current,
                EPC.measured_cumulative_energy,
                EPC.measured_cumulative_energy_reversed,
            })
            for epc, r in res.items():
                print(f'epc: {epc.name}, result: {r}')
            print('----')
            time.sleep(5)

            print('---- closing the session ----')
            exit_code = 0
            break
    except (momonga.MomongaSkScanFailure,
            momonga.MomongaSkJoinFailure,
            momonga.MomongaNeedToReopen,
            ):
        time.sleep(60)
        continue
    except Exception as e:
        print(traceback.format_exc(), file=sys.stderr)
        break

exit(exit_code)



import logging
import threading
import queue
import time

from typing import Self

from .momonga_exception import (MomongaSkScanFailure,
                                MomongaSkJoinFailure,
                                MomongaNeedToReopen,
                                MomongaSkCommandExecutionFailure,
)
from .momonga_sk_wrapper import MomongaSkWrapper
from .momonga_sk_wrapper import logger as sk_wrapper_logger

logger = logging.getLogger(__name__)


class MomongaSessionManager:
    def __init__(self,
                 rbid: str,
                 pwd: str,
                 dev: str,
                 baudrate: int = 115200,
                 reset_dev: bool = True,
                ) -> None:
        self.dev = dev
        self.baudrate = baudrate
        self.rbid = rbid
        self.pwd = pwd
        self.reset_dev = reset_dev

        # the following value will be set a pyserial object.
        self.skw = MomongaSkWrapper(dev, baudrate)

        # the following values will be set by open() with skscan().
        self.smart_meter_mac = None
        self.smart_meter_addr = None
        self.channel = None
        self.pan_id = None

        # the following values will be set by open() with skjoin().
        self.session_established = False
        self.receiver_th = None
        self.receiver_exception = None
        self.xmit_restriction_cnt = 0
        self.xmit_lock = threading.Lock()
        self.rejoin_lock = threading.Lock()

        self.pkt_sbsc_q = queue.Queue()
        self.recv_q = queue.Queue()
        self.xmit_q = queue.Queue()

    def __enter__(self) -> Self:
        return self.open()

    def __exit__(self, type, value, traceback) -> None:
        self.close()

    def open(self) -> Self:
        logger.info('Opening a Momonga session...')
        try:
            self.skw.open()

            if self.reset_dev is True:
                # to reset the specified wi-sun module.
                self.skw.skreset()

            # to disable echoback.
            # self.sksreg('SFE', '0')

            # to show the rssi of the received packets.
            self.skw.sksreg('SA2', '1')

            # scanning a PAN from here.
            # to set a route b id.
            self.skw.sksetrbid(self.rbid)
            # to set a password.
            self.skw.sksetpwd(self.pwd)
            logger.info('The Route-B ID and the password were registered.')
            try:
                logger.info('Scanning PAN channels...')
                scan_res = self.skw.skscan()
                logger.info('A PAN was found.')
            except MomongaSkScanFailure:
                logger.error('Gave up to find a PAN. Check the device location and Route-B ID. Then try again.')
                raise MomongaSkScanFailure('Gave up to find a PAN. Check the device location and Route-B ID. Then try again.')
            self.smart_meter_mac = scan_res.mac_addr
            self.channel = scan_res.channel
            self.pan_id = scan_res.pan_id
            # converting mac addr to ip6 addr.
            self.smart_meter_addr = self.skw.skll64(scan_res.mac_addr).ip6_addr

            # joining a PAN from here.
            logger.info('Joining the PAN...')
            # to set a channel.
            self.skw.sksreg('S2', self.channel)
            # to set a pan id.
            self.skw.sksreg('S3', self.pan_id)
            # to establish a pana session.
            try:
                self.skw.skjoin(self.smart_meter_addr)
                self.session_established = True
                logger.info('A PANA session has been established.')
            except MomongaSkJoinFailure:
                logger.error('Gave up to establish a PANA session. Check the Route-B ID and password. Then try again.')
                raise MomongaSkJoinFailure('Gave up to establish a PANA session. Check the Route-B ID and password. Then try again.')

            while not self.pkt_sbsc_q.empty():
                self.pkt_sbsc_q.get()
            while not self.recv_q.empty():
                self.recv_q.get()
            while not self.xmit_q.empty():
                self.xmit_q.get()

            self.receiver_th = threading.Thread(target=self.receiver, daemon=True)
            self.skw.subscribers.update({'pkt_sbsc_q': self.pkt_sbsc_q})
            self.receiver_th.start()

            logger.info('A Momonga session is open.')
            return self
        except Exception as e:
            logger.error('Could not open a Momonga session. %s: %s' % (type(e).__name__, e))
            self.close()
            raise e

    def close(self) -> None:
        logger.info('Closing the Momonga session...')

        if self.rejoin_lock.acquire(timeout=120) is False:
            logger.warning('Failed to acquire "rejoin_lock".')

        if self.session_established is True:
            try:
                self.session_established = False
                logger.info('Terminating the PANA session...')
                self.skw.skterm()
            except Exception as e:
                logger.warning('Failed to terminate the PANA session. %s: %s' % (type(e).__name__, e))
            finally:
                self.rejoin_lock.release()
        else:
            self.rejoin_lock.release()

        if self.receiver_th is not None:
            if self.receiver_th.is_alive():
                self.pkt_sbsc_q.put('__CLOSE__')  # to close the receiver thread.
                self.receiver_th.join()
            self.receiver_th = None

        if self.skw.subscribers.get('pkt_sbsc_q') is not None:
            self.skw.subscribers.pop('pkt_sbsc_q')

        self.unrestrict_to_xmit(force=True)

        assert self.xmit_lock.locked() is False, '"xmit_lock" is unexpectedly locked.'
        assert self.rejoin_lock.locked() is False, '"rejoin_lock" is unexpectedly locked.'

        self.skw.close()
        logger.info('The Momonga session is closed.')

    def receiver(self) -> None:
        logger.debug('A packet receiver has been started.')
        try:
            while True:
                res = self.pkt_sbsc_q.get()
                if res == '__CLOSE__':
                    break

                if not (res.startswith('EVENT') or res.startswith('ERXUDP')):
                    # messages that do not need to be handled will be discarded.
                    continue

                if res.startswith('EVENT 29'):
                    logger.debug('The PANA session lifetime has been expired.')
                    self.restrict_to_xmit()
                elif res.startswith('EVENT 24'):
                    logger.warning('Could not rejoin the PAN.')
                    self.rejoin_lock.acquire()
                    if self.session_established is True:
                        self.session_established = False
                        try:
                            self.skw.skjoin(self.smart_meter_addr)
                        except MomongaSkJoinFailure as e:
                            logger.error('%s Close Momonga and open it again.' % (e))
                            raise MomongaNeedToReopen('%s Close Momonga and open it again.' % (e))
                        finally:
                            self.rejoin_lock.release()
                    else:
                        self.rejoin_lock.release()
                elif res.startswith('EVENT 25'):
                    logger.debug('Successfully rejoined the PAN.')
                    self.session_established = True
                    self.unrestrict_to_xmit()
                elif res.startswith('EVENT 32'):
                    logger.warning('The transmission rate limit has been exceeded.')
                    self.restrict_to_xmit()
                elif res.startswith('EVENT 33'):
                    logger.debug('The transmission rate limit has been released.')
                    self.unrestrict_to_xmit()
                elif res.startswith('EVENT 27'):
                    self.restrict_to_xmit()
                    logger.debug('The PANA session has been closed successfully.')
                elif res.startswith('EVENT 28'):  # there was no session to close.
                    self.restrict_to_xmit()
                    logger.warning('There was no PANA session to close.')
                elif res.startswith("EVENT 21") or res.startswith("EVENT 02"):
                    if self.is_restricted_to_xmit() is False:
                        self.recv_q.put(res)
                elif res.startswith("ERXUDP"):
                    self.recv_q.put(res)
                else:
                    # other events that do not need to be handled will be discarded.
                    continue
        except Exception as e:
            logger.error('An exception was raised from the receiver thread. %s: %s' % (type(e).__name__, e))
            self.receiver_exception = e

        logger.debug('The packet receiver has been stopped.')

    def xmitter(self,
                data: bytes,
               ) -> None:
        retry_to_xmit = 3
        retry_to_acquire_xmit_lock = 60
        xmitted = False
        for _ in range(retry_to_xmit):
            logger.debug('Trying to acquire "xmit_lock".')
            locked = False
            for r in range(retry_to_acquire_xmit_lock):
                locked = self.xmit_lock.acquire(timeout=60)
                if locked is False:
                    logger.warning('Could not acquire "xmit_lock". (%d/%d)' % (r + 1, retry_to_acquire_xmit_lock))
                    if self.receiver_exception is not None:
                        logger.error('Got an exception from the receiver thread. %s: %s' % (type(self.receiver_exception).__name__, self.receiver_exception))
                        raise MomongaNeedToReopen('Got an exception from the receiver thread. %s: %s' % (type(self.receiver_exception).__name__, self.receiver_exception))
                else:
                    break

            if locked is False:
                logger.error('Transmission rights could not be acquired. Close Momonga and open it again.')
                raise MomongaNeedToReopen('Transmission rights could not be acquired. Close Momonga and open it again.')
            else:
                logger.debug('Acquired "xmit_lock".')

            assert self.session_established is not False, 'Tried to transmit a packet, but no PANA session was established.'

            try:
                self.skw.sksendto(self.smart_meter_addr, data)
                xmitted = True
                break
            except MomongaSkCommandExecutionFailure as e:
                logger.warning('Failed to transmit a packet: %s' % (e))
            except Exception as e:
                logger.warning('An error occurred to transmit a packet. %s: %s' % (type(e).__name__, e))
            finally:
                self.xmit_lock.release()
            time.sleep(3)
        if xmitted is False:
            logger.error('Could not transmit a packet. Close Momonga and open it again.')
            raise MomongaNeedToReopen('Could not transmit a packet. Close Momonga and open it again.')

    def restrict_to_xmit(self) -> None:
        self.xmit_restriction_cnt += 1
        logger.debug('The counter for the restriction was incremented: %d' % (self.xmit_restriction_cnt))

        assert self.xmit_restriction_cnt <= 2, 'The critical section counter for data transmission is inconsistent: Too big than expected.'

        if self.xmit_restriction_cnt == 1:
            logger.debug('Trying to restrict data transmission.')
            self.xmit_lock.acquire()
            logger.debug('Data transmission is being restricted.')

    def unrestrict_to_xmit(self,
                           force=False,
                          ) -> None:
        if force is True:
            self.xmit_restriction_cnt = 0
            logger.debug('The counter for the restriction was forcibly set to zero.')
        else:
            self.xmit_restriction_cnt -= 1
            logger.debug('The counter for the restriction was decremented: %d' % (self.xmit_restriction_cnt))

        assert self.xmit_restriction_cnt >= 0, 'The critical section counter for data transmit is inconsistent: Too small than expected.'

        if self.xmit_restriction_cnt == 0:
            try:
                self.xmit_lock.release()
            except RuntimeError:
                pass  # xmit_lock is likely already unlocked.

            logger.debug('Data transmission is being unrestricted.')

    def is_restricted_to_xmit(self) -> bool:
        if self.xmit_restriction_cnt == 0:
            return False
        else:
            return True



import datetime
import enum
import time
import queue
import inspect
import logging

from typing import TypedDict, Any, Self

from .momonga_exception import (MomongaResponseNotExpected,
                                MomongaResponseNotPossible,
                                MomongaNeedToReopen,
                                MomongaRuntimeError)
from .momonga_response import SkEventRxUdp
from .momonga_session_manager import MomongaSessionManager
from .momonga_session_manager import logger as session_manager_logger
from .momonga_sk_wrapper import logger as sk_wrapper_logger

logger = logging.getLogger(__name__)


class EchonetServiceCode(enum.IntEnum):
    set_c: int = 0x61
    get: int = 0x62


class EchonetPropertyCode(enum.IntEnum):
    operation_status: int = 0x80
    installation_location: int = 0x81
    standard_version_information: int = 0x82
    fault_status: int = 0x88
    manufacturer_code: int = 0x8A
    serial_number: int = 0x8D
    current_time_setting: int = 0x97
    current_date_setting: int = 0x98
    properties_for_status_notification: int = 0x9D
    properties_to_set_values: int = 0x9E
    properties_to_get_values: int = 0x9F
    route_b_id: int = 0xC0
    one_minute_measured_cumulative_energy: int = 0xD0
    coefficient_for_cumulative_energy: int = 0xD3
    number_of_effective_digits_for_cumulative_energy: int = 0xD7
    measured_cumulative_energy: int = 0xE0
    measured_cumulative_energy_reversed: int = 0xE3
    unit_for_cumulative_energy: int = 0xE1
    historical_cumulative_energy_1: int = 0xE2
    historical_cumulative_energy_1_reversed: int = 0xE4
    day_for_historical_data_1: int = 0xE5
    instantaneous_power: int = 0xE7
    instantaneous_current: int = 0xE8
    cumulative_energy_measured_at_fixed_time: int = 0xEA
    cumulative_energy_measured_at_fixed_time_reversed: int = 0xEB
    historical_cumulative_energy_2: int = 0xEC
    time_for_historical_data_2: int = 0xED
    historical_cumulative_energy_3: int = 0xEE
    time_for_historical_data_3: int = 0xEF


class EchonetProperty:
    def __init__(self,
                 epc: EchonetPropertyCode | int,
                 ) -> None:
        self.epc = epc


class EchonetPropertyWithData:
    def __init__(self,
                 epc: EchonetPropertyCode | int,
                 edt: bytes | None = None,
                 ) -> None:
        self.epc = epc
        self.edt = edt


class EchonetDataParser:
    @classmethod
    def parse_operation_status(cls, edt: bytes) -> bool | None:
        status = int.from_bytes(edt, 'big')
        if status == 0x30:  # turned on
            status = True
        elif status == 0x31:  # turned off
            status = False
        else:
            status = None  # unknown

        return status

    @classmethod
    def parse_installation_location(cls, edt: bytes) -> str:
        code = edt[0]
        if code == 0x00:
            location = 'location not set'
        elif code == 0x01:
            location = 'location information: ' + edt[1:].hex()
        elif 0x02 <= code <= 0x07:  # reserved for future use
            location = 'not implemented'
        elif 0x08 <= code <= 0x7F:
            location_map = {
                1: 'living room',
                2: 'dining room',
                3: 'kitchen',
                4: 'bathroom',
                5: 'toilet',
                6: 'washroom',
                7: 'hallway',
                8: 'room',
                9: 'stairs',
                10: 'entrance',
                11: 'storage room',
                12: 'garden/perimeter',
                13: 'garage',
                14: 'veranda',
                15: 'other',
            }
            location_code = code >> 3
            location = location_map[location_code]
            location += ' ' + str(code & 0x07)
        elif 0x80 <= code <= 0xFE:  # reserved for future use
            location = 'not implemented'
        elif code == 0xFF:
            location = 'location not fixed'
        else:
            location = 'unknown'

        return location

    @classmethod
    def parse_standard_version_information(cls, edt: bytes) -> str:
        version = ''
        if edt[0] > 0:
            version += chr(edt[0])
        if edt[1] > 0:
            version += chr(edt[1])
        return version + chr(edt[2]) + '.' + str(edt[3])

    @classmethod
    def parse_fault_status(cls, edt: bytes) -> bool:
        status_code = int.from_bytes(edt, 'big')
        if status_code == 0x41:
            status = True  # fault occurred
        elif status_code == 0x42:
            status = False  # no fault occurred
        else:
            status = None  # unknown

        return status

    @classmethod
    def parse_manufacturer_code(cls, edt: bytes) -> bytes:
        return edt

    @classmethod
    def parse_serial_number(cls, edt: bytes) -> str:
        return edt.decode()

    @classmethod
    def parse_current_time_setting(cls, edt: bytes) -> datetime.time:
        hour = edt[0]
        minute = edt[1]
        return datetime.time(hour=hour, minute=minute, second=0)

    @classmethod
    def parse_current_date_setting(cls, edt: bytes) -> datetime.date:
        year = int.from_bytes(edt[0:2], 'big')
        month = edt[2]
        day = edt[3]
        return datetime.date(year=year, month=month, day=day)

    @classmethod
    def parse_property_map(cls, edt: bytes) -> set[EchonetPropertyCode | int]:
        num_of_properties = edt[0]
        property_map = edt[1:]
        properties = set()
        if num_of_properties < 16:
            for prop_code in property_map:
                try:
                    prop_code = EchonetPropertyCode(prop_code)
                except ValueError:
                    pass

                properties.add(prop_code)
        else:
            for i in range(len(property_map)):
                b = property_map[i]
                for j in range(8):
                    if b & 1 << j:
                        prop_code = (j + 0x08 << 4) + i
                        try:
                            prop_code = EchonetPropertyCode(prop_code)
                        except ValueError:
                            pass

                        properties.add(prop_code)

        return properties

    @classmethod
    def parse_route_b_id(cls, edt: bytes) -> dict[str, bytes]:
        manufacturer_code = edt[1:4]
        authentication_id = edt[4:]
        return {'manufacturer code': manufacturer_code, 'authentication id': authentication_id}

    @classmethod
    def parse_one_minute_measured_cumulative_energy(
            cls,
            edt: bytes,
            energy_unit: int | float,
            energy_coefficient: int,
    ) -> dict[str, datetime.datetime | dict[str, int | float | None]]:
        timestamp = datetime.datetime(int.from_bytes(edt[0:2], 'big'),
                                      edt[2], edt[3], edt[4], edt[5], edt[6])

        normal_direction_energy = int.from_bytes(edt[7:11], 'big')
        if normal_direction_energy == 0xFFFFFFFE:
            normal_direction_energy = None
        else:
            normal_direction_energy *= energy_unit
            normal_direction_energy *= energy_coefficient

        reverse_direction_energy = int.from_bytes(edt[11:15], 'big')
        if reverse_direction_energy == 0xFFFFFFFE:
            reverse_direction_energy = None
        else:
            reverse_direction_energy *= energy_unit
            reverse_direction_energy *= energy_coefficient

        return {'timestamp': timestamp,
                'cumulative energy': {'normal direction': normal_direction_energy,
                                      'reverse direction': reverse_direction_energy}}

    @classmethod
    def parse_coefficient_for_cumulative_energy(cls, edt: bytes) -> int:
        coefficient = int.from_bytes(edt, 'big')
        return coefficient

    @classmethod
    def parse_number_of_effective_digits_for_cumulative_energy(cls, edt: bytes) -> int:
        digits = int.from_bytes(edt, 'big')
        return digits

    @classmethod
    def parse_measured_cumulative_energy(
            cls,
            edt: bytes,
            energy_unit: int | float,
            energy_coefficient: int,
    ) -> int | float:
        cumulative_energy = int.from_bytes(edt, 'big')
        cumulative_energy *= energy_unit
        cumulative_energy *= energy_coefficient
        return cumulative_energy

    @classmethod
    def parse_unit_for_cumulative_energy(cls, edt: bytes) -> int | float:
        unit_index = int.from_bytes(edt, 'big')
        unit_map = {0x00: 1,
                    0x01: 0.1,
                    0x02: 0.01,
                    0x03: 0.001,
                    0x04: 0.0001,
                    0x0A: 10,
                    0x0B: 100,
                    0x0C: 1000,
                    0x0D: 10000}
        unit = unit_map.get(unit_index)
        if unit is None:
            raise MomongaRuntimeError('Obtained unit for cumulative energy (%X) is not defined.' % unit_index)

        return unit

    @classmethod
    def parse_historical_cumulative_energy_1(
            cls,
            edt: bytes,
            energy_unit: int | float,
            energy_coefficient: int,
    ) -> list[dict[str, datetime.datetime | int | float | None]]:
        day = int.from_bytes(edt[0:2], 'big')
        timestamp = datetime.datetime.combine(datetime.date.today(), datetime.datetime.min.time())
        timestamp -= datetime.timedelta(days=day)
        energy_data_points = edt[2:]
        historical_cumulative_energy = []
        for i in range(48):
            j = i * 4
            cumulative_energy = int.from_bytes(energy_data_points[j:j + 4], 'big')
            if cumulative_energy == 0xFFFFFFFE:
                cumulative_energy = None
            else:
                cumulative_energy *= energy_unit
                cumulative_energy *= energy_coefficient
            historical_cumulative_energy.append({'timestamp': timestamp, 'cumulative energy': cumulative_energy})
            timestamp += datetime.timedelta(minutes=30)

        return historical_cumulative_energy

    @classmethod
    def parse_day_for_historical_data_1(cls, edt: bytes) -> int:
        day = int.from_bytes(edt, 'big')
        return day

    @classmethod
    def parse_instantaneous_power(cls, edt: bytes) -> float:
        power = int.from_bytes(edt, 'big', signed=True)
        return power

    @classmethod
    def parse_instantaneous_current(cls, edt: bytes) -> dict[str, float]:
        r_phase_current = int.from_bytes(edt[0:2], 'big', signed=True)
        t_phase_current = int.from_bytes(edt[2:4], 'big', signed=True)
        r_phase_current *= 0.1  # to Ampere
        t_phase_current *= 0.1  # to Ampere
        return {'r phase current': r_phase_current, 't phase current': t_phase_current}

    @classmethod
    def parse_cumulative_energy_measured_at_fixed_time(
            cls,
            edt: bytes,
            energy_unit: int | float,
            energy_coefficient: int,
    ) -> dict[str, datetime.datetime | int | float]:
        timestamp = datetime.datetime(int.from_bytes(edt[0:2], 'big'),
                                      edt[2], edt[3], edt[4], edt[5], edt[6])
        cumulative_energy = int.from_bytes(edt[7:], 'big')
        cumulative_energy *= energy_unit
        cumulative_energy *= energy_coefficient
        return {'timestamp': timestamp, 'cumulative energy': cumulative_energy}

    @classmethod
    def parse_historical_cumulative_energy_2(
            cls,
            edt: bytes,
            energy_unit: int | float,
            energy_coefficient: int,
    ) -> list[dict[str, datetime.datetime |
                         dict[str, int | float | None]]]:
        year = int.from_bytes(edt[0:2], 'big')
        num_of_data_points = edt[6]
        energy_data_points = edt[7:]
        timestamp = datetime.datetime(year, edt[2], edt[3], edt[4], edt[5])
        historical_cumulative_energy = []
        for i in range(num_of_data_points):
            j = i * 8
            normal_direction_energy = int.from_bytes(energy_data_points[j:j + 4], 'big')
            if normal_direction_energy == 0xFFFFFFFE:
                normal_direction_energy = None
            else:
                normal_direction_energy *= energy_unit
                normal_direction_energy *= energy_coefficient

            reverse_direction_energy = int.from_bytes(energy_data_points[j + 4:j + 8], 'big')
            if reverse_direction_energy == 0xFFFFFFFE:
                reverse_direction_energy = None
            else:
                reverse_direction_energy *= energy_unit
                reverse_direction_energy *= energy_coefficient

            historical_cumulative_energy.append(
                {'timestamp': timestamp,
                 'cumulative energy': {'normal direction': normal_direction_energy,
                                       'reverse direction': reverse_direction_energy}})
            timestamp -= datetime.timedelta(minutes=30)

        return historical_cumulative_energy

    @classmethod
    def parse_time_for_historical_data_2(cls, edt: bytes) -> dict[str, datetime.datetime | None | int]:
        year = int.from_bytes(edt[0:2], 'big')
        if year == 0xFFFF:
            timestamp = None
        else:
            timestamp = datetime.datetime(year, edt[2], edt[3], edt[4], edt[5])

        num_of_data_points = edt[6]
        return {'timestamp': timestamp,
                'number of data points': num_of_data_points}

    @classmethod
    def parse_historical_cumulative_energy_3(
            cls,
            edt: bytes,
            energy_unit: int | float,
            energy_coefficient: int,
    ) -> list[dict[str, datetime.datetime |
                        dict[str, dict[str, int | float | None]]]]:
        year = int.from_bytes(edt[0:2], 'big')
        num_of_data_points = edt[6]
        energy_data_points = edt[7:]
        timestamp = datetime.datetime(year, edt[2], edt[3], edt[4], edt[5])
        historical_cumulative_energy = []
        for i in range(num_of_data_points):
            j = i * 8
            normal_direction_energy = int.from_bytes(energy_data_points[j:j + 4], 'big')
            if normal_direction_energy == 0xFFFFFFFE:
                normal_direction_energy = None
            else:
                normal_direction_energy *= energy_unit
                normal_direction_energy *= energy_coefficient

            reverse_direction_energy = int.from_bytes(energy_data_points[j + 4:j + 8], 'big')
            if reverse_direction_energy == 0xFFFFFFFE:
                reverse_direction_energy = None
            else:
                reverse_direction_energy *= energy_unit
                reverse_direction_energy *= energy_coefficient

            historical_cumulative_energy.append(
                {'timestamp': timestamp,
                 'cumulative energy': {'normal direction': normal_direction_energy,
                                       'reverse direction': reverse_direction_energy}})
            timestamp -= datetime.timedelta(minutes=1)

        return historical_cumulative_energy

    @classmethod
    def parse_time_for_historical_data_3(cls, edt: bytes) -> dict[str, datetime.datetime | None | int]:
        year = int.from_bytes(edt[0:2], 'big')
        if year == 0xFFFF:
            timestamp = None
        else:
            timestamp = datetime.datetime(year, edt[2], edt[3], edt[4], edt[5])

        num_of_data_points = edt[6]
        return {'timestamp': timestamp,
                'number of data points': num_of_data_points}


parser_map: dict[EchonetPropertyCode, callable] = {
    EchonetPropertyCode.operation_status: EchonetDataParser.parse_operation_status,
    EchonetPropertyCode.installation_location: EchonetDataParser.parse_installation_location,
    EchonetPropertyCode.standard_version_information: EchonetDataParser.parse_standard_version_information,
    EchonetPropertyCode.fault_status: EchonetDataParser.parse_fault_status,
    EchonetPropertyCode.manufacturer_code: EchonetDataParser.parse_manufacturer_code,
    EchonetPropertyCode.serial_number: EchonetDataParser.parse_serial_number,
    EchonetPropertyCode.current_time_setting: EchonetDataParser.parse_current_time_setting,
    EchonetPropertyCode.current_date_setting: EchonetDataParser.parse_current_date_setting,
    EchonetPropertyCode.properties_for_status_notification: EchonetDataParser.parse_property_map,
    EchonetPropertyCode.properties_to_set_values: EchonetDataParser.parse_property_map,
    EchonetPropertyCode.properties_to_get_values: EchonetDataParser.parse_property_map,
    EchonetPropertyCode.route_b_id: EchonetDataParser.parse_route_b_id,
    EchonetPropertyCode.one_minute_measured_cumulative_energy: EchonetDataParser.parse_one_minute_measured_cumulative_energy,
    EchonetPropertyCode.coefficient_for_cumulative_energy: EchonetDataParser.parse_coefficient_for_cumulative_energy,
    EchonetPropertyCode.number_of_effective_digits_for_cumulative_energy: EchonetDataParser.parse_number_of_effective_digits_for_cumulative_energy,
    EchonetPropertyCode.measured_cumulative_energy: EchonetDataParser.parse_measured_cumulative_energy,
    EchonetPropertyCode.measured_cumulative_energy_reversed: EchonetDataParser.parse_measured_cumulative_energy,
    EchonetPropertyCode.unit_for_cumulative_energy: EchonetDataParser.parse_unit_for_cumulative_energy,
    EchonetPropertyCode.historical_cumulative_energy_1: EchonetDataParser.parse_historical_cumulative_energy_1,
    EchonetPropertyCode.historical_cumulative_energy_1_reversed: EchonetDataParser.parse_historical_cumulative_energy_1,
    EchonetPropertyCode.day_for_historical_data_1: EchonetDataParser.parse_day_for_historical_data_1,
    EchonetPropertyCode.instantaneous_power: EchonetDataParser.parse_instantaneous_power,
    EchonetPropertyCode.instantaneous_current: EchonetDataParser.parse_instantaneous_current,
    EchonetPropertyCode.cumulative_energy_measured_at_fixed_time: EchonetDataParser.parse_cumulative_energy_measured_at_fixed_time,
    EchonetPropertyCode.cumulative_energy_measured_at_fixed_time_reversed: EchonetDataParser.parse_cumulative_energy_measured_at_fixed_time,
    EchonetPropertyCode.historical_cumulative_energy_2: EchonetDataParser.parse_historical_cumulative_energy_2,
    EchonetPropertyCode.time_for_historical_data_2: EchonetDataParser.parse_time_for_historical_data_2,
    EchonetPropertyCode.historical_cumulative_energy_3: EchonetDataParser.parse_historical_cumulative_energy_3,
    EchonetPropertyCode.time_for_historical_data_3: EchonetDataParser.parse_time_for_historical_data_3,
}


class EchonetDataBuilder:
    @classmethod
    def build_edata_to_set_day_for_historical_data_1(cls, day: int = 0) -> bytes:
        if day < 0 or day > 99:
            raise ValueError('The parameter "day" must be between 0 and 99.')

        return day.to_bytes(1, 'big')

    @classmethod
    def build_edata_to_set_time_for_historical_data_2(cls,
                                                      timestamp: datetime.datetime,
                                                      num_of_data_points: int = 12,
                                                      ) -> bytes:
        if num_of_data_points < 1 or num_of_data_points > 12:
            raise ValueError('The parameter "num_of_data_points" must be between 1 and 12.')

        if timestamp.year < 1 or timestamp.year > 9999:
            raise ValueError('The year specified by the parameter "timestamp" must be between 1 and 9999.')

        year = timestamp.year.to_bytes(2, 'big')
        month = timestamp.month.to_bytes(1, 'big')
        day = timestamp.day.to_bytes(1, 'big')
        hour = timestamp.hour.to_bytes(1, 'big')
        if 0 <= timestamp.minute < 30:
            minute = 0
        else:
            minute = 30

        minute = minute.to_bytes(1, 'big')
        num_of_data_points = num_of_data_points.to_bytes(1, 'big')
        return year + month + day + hour + minute + num_of_data_points

    @classmethod
    def build_edata_to_set_time_for_historical_data_3(cls,
                                                      timestamp: datetime.datetime,
                                                      num_of_data_points: int = 10,
                                                      ) -> bytes:
        if num_of_data_points < 1 or num_of_data_points > 10:
            raise ValueError('The parameter "num_of_data_points" must be between 1 and 10.')

        if timestamp.year < 1 or timestamp.year > 9999:
            raise ValueError('The year specified by the parameter "timestamp" must be between 1 and 9999.')

        year = timestamp.year.to_bytes(2, 'big')
        month = timestamp.month.to_bytes(1, 'big')
        day = timestamp.day.to_bytes(1, 'big')
        hour = timestamp.hour.to_bytes(1, 'big')
        minute = timestamp.minute.to_bytes(1, 'big')
        num_of_data_points = num_of_data_points.to_bytes(1, 'big')
        return year + month + day + hour + minute + num_of_data_points


class Momonga:
    def __init__(self,
                 rbid: str,
                 pwd: str,
                 dev: str,
                 baudrate: int = 115200,
                 reset_dev: bool = True,
                 ) -> None:
        self.xmit_retries: int = 12
        self.recv_timeout: int | float = 12
        self.internal_xmit_interval: int | float = 5
        self.transaction_id: int = 0
        self.energy_unit: int | float = 1
        self.energy_coefficient: int = 1
        self.is_open: bool = False
        self.session_manager = MomongaSessionManager(rbid, pwd, dev, baudrate, reset_dev)

    def __init_energy_unit(self) -> None:
        logger.debug('Initializing the energy unit and coefficient.')
        self.energy_unit = self.get_unit_for_cumulative_energy()
        try:
            self.energy_coefficient = self.get_coefficient_for_cumulative_energy()
            time.sleep(self.internal_xmit_interval)
        except MomongaResponseNotPossible:  # due to the property 0xD3 is optional.
            self.energy_coefficient = 1
        time.sleep(self.internal_xmit_interval)

    def __enter__(self) -> Self:
        return self.open()

    def __exit__(self, type, value, traceback) -> None:
        self.close()

    def open(self) -> Self:
        logger.info('Opening Momonga.')
        self.session_manager.open()
        time.sleep(self.internal_xmit_interval)
        self.is_open = True
        self.__init_energy_unit()
        logger.info('Momonga is open.')
        return self

    def close(self) -> None:
        logger.info('Closing Momonga.')
        self.is_open = False
        self.session_manager.close()
        logger.info('Momonga is closed.')

    def __get_transaction_id(self) -> int:
        self.transaction_id += 1
        return self.transaction_id

    @staticmethod
    def __build_request_header(tid: int, esv: EchonetServiceCode) -> bytes:
        ehd = b'\x10\x81'  # echonet lite edata format 1
        tid = tid.to_bytes(4, 'big')[-2:]
        seoj = b'\x05\xFF\x01'  # controller class
        deoj = b'\x02\x88\x01'  # low-voltage smart electric energy meter class
        esv = esv.to_bytes(1, 'big')
        return ehd + tid + seoj + deoj + esv

    def __build_request_payload_with_data(self,
                                          tid: int,
                                          esv: EchonetServiceCode,
                                          properties_with_data: list[EchonetPropertyWithData],
                                          ) -> bytes:
        header = self.__build_request_header(tid, esv)
        opc = len(properties_with_data).to_bytes(1, 'big')
        payload = header + opc
        for pd in properties_with_data:
            epc = pd.epc.to_bytes(1, 'big')
            pdc = len(pd.edt).to_bytes(1, 'big')
            edt = pd.edt
            payload += epc + pdc + edt

        return payload

    def __build_request_payload(self,
                                tid: int,
                                esv: EchonetServiceCode,
                                properties: list[EchonetProperty],
                                ) -> bytes:
        header = self.__build_request_header(tid, esv)  # get
        opc = len(properties).to_bytes(1, 'big')
        payload = header + opc
        for p in properties:
            epc = p.epc.to_bytes(1, 'big')
            pdc = b'\x00'
            payload += epc + pdc

        return payload

    @staticmethod
    def __extract_response_payload(data: bytes,
                                   tid: int,
                                   req_properties: list[EchonetPropertyWithData] | list[EchonetProperty],
                                   ) -> list[EchonetPropertyWithData]:
        ehd = data[0:2]
        if ehd != b'\x10\x81':  # echonet lite edata format 1
            raise MomongaResponseNotExpected('The data format is not ECHONET Lite EDATA format 1')

        if data[2:4] != tid.to_bytes(4, 'big')[-2:]:
            raise MomongaResponseNotExpected('The transaction ID does not match.')

        seoj = data[4:7]
        if seoj != b'\x02\x88\x01':  # low-voltage smart electric energy meter class
            raise MomongaResponseNotExpected('The source is not a smart meter.')

        deoj = data[7:10]
        if deoj != b'\x05\xFF\x01':  # controller class
            raise MomongaResponseNotExpected('The destination is not a controller.')

        esv = data[10]
        if 0x50 <= esv <= 0x5F:
            raise MomongaResponseNotPossible('The target smart meter could not respond. ESV: %X' % esv)

        opc = data[11]
        req_opc = len(req_properties)
        if opc != req_opc:
            raise MomongaResponseNotExpected(
                'Unexpected packet format. OPC is expected %s but %d was set.' % (req_opc, opc))

        properties = []
        cur = 12
        for rp in req_properties:
            try:
                epc = EchonetPropertyCode(data[cur])
            except ValueError:
                epc = data[cur]

            if epc != rp.epc:
                raise MomongaResponseNotExpected('The property code does not match. EPC: %X' % rp.epc)

            cur += 1
            pdc = data[cur]
            cur += 1
            if pdc == 0:
                edt = None
            else:
                edt_from = cur
                cur += pdc
                edt = data[edt_from:cur]

            properties.append(EchonetPropertyWithData(epc, edt))

        return properties

    def __request(self,
                  esv: EchonetServiceCode,
                  req_properties: list[EchonetPropertyWithData] | list[EchonetProperty],
                  ) -> list[EchonetPropertyWithData]:
        logger.debug('Checking if Momonga is open: is_open=%s', self.is_open)
        if self.is_open is not True:
            raise RuntimeError('Momonga is not open.')

        tid = self.__get_transaction_id()
        if esv == EchonetServiceCode.set_c:
            tx_payload = self.__build_request_payload_with_data(tid, esv, req_properties)
        elif esv == EchonetServiceCode.get:
            tx_payload = self.__build_request_payload(tid, esv, req_properties)
        else:
            raise MomongaRuntimeError('Unsupported service code.')

        while not self.session_manager.recv_q.empty():
            self.session_manager.recv_q.get()  # drops stored data

        for _ in range(self.xmit_retries):
            self.session_manager.xmitter(tx_payload)
            while True:
                try:
                    res = self.session_manager.recv_q.get(timeout=self.recv_timeout)
                except queue.Empty:
                    logger.warning('The request for transaction id "%04X" timed out.' % tid)
                    break  # to rexmit the request.

                # messages of event types 21, 02, and received udp payloads will only be delivered.
                if res.startswith('EVENT 21'):
                    param = res.split()[-1]
                    if param == '00':
                        logger.info('Successfully transmitted a request packet for transaction id "%04X".' % tid)
                        continue
                    elif param == '01':
                        logger.info('Retransmitting the request packet for transaction id "%04X".' % tid)
                        time.sleep(self.internal_xmit_interval)
                        break  # to rexmit the request.
                    elif param == '02':
                        logger.info('Transmitting neighbor solicitation packets.')
                        continue
                    else:
                        logger.debug('A message for event 21 with an unknown parameter "%s" will be ignored.' % param)
                        continue
                elif res.startswith('EVENT 02'):
                    logger.info('Received a neighbor advertisement packet.')
                    continue
                elif res.startswith('ERXUDP'):
                    udp_pkt = SkEventRxUdp([res], self.session_manager.skw.device_type)
                    if not (udp_pkt.src_port == udp_pkt.dst_port == 0x0E1A):
                        continue
                    elif udp_pkt.side:
                        continue
                    elif udp_pkt.src_addr != self.session_manager.smart_meter_addr:
                        continue

                    try:
                        res_properties = self.__extract_response_payload(udp_pkt.data, tid, req_properties)
                    except MomongaResponseNotExpected:
                        continue

                    logger.info('Successfully received a response packet for transaction id "%04X".' % tid)
                    return res_properties
                else:
                    # this line should never be reached.
                    continue
        logger.error('Gave up to obtain a response for transaction id "%04X". Close Momonga and open it again.' % tid)
        raise MomongaNeedToReopen('Gave up to obtain a response for transaction id "%04X".'
                                  ' Close Momonga and open it again.' % tid)

    def __request_to_set(self,
                         properties_with_data: list[EchonetPropertyWithData]
                         ) -> None:
        self.__request(EchonetServiceCode.set_c, properties_with_data)

    def __request_to_get(self,
                         properties: list[EchonetProperty],
                         ) -> list[EchonetPropertyWithData]:
        return self.__request(EchonetServiceCode.get, properties)

    def get_operation_status(self) -> bool | None:
        req = EchonetProperty(EchonetPropertyCode.operation_status)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_operation_status(res.edt)

    def get_installation_location(self) -> str:
        req = EchonetProperty(EchonetPropertyCode.installation_location)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_installation_location(res.edt)

    def get_standard_version(self) -> str:
        req = EchonetProperty(EchonetPropertyCode.standard_version_information)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_standard_version_information(res.edt)

    def get_fault_status(self) -> bool | None:
        req = EchonetProperty(EchonetPropertyCode.fault_status)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_fault_status(res.edt)

    def get_manufacturer_code(self) -> bytes:
        req = EchonetProperty(EchonetPropertyCode.manufacturer_code)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_manufacturer_code(res.edt)

    def get_serial_number(self) -> str:
        req = EchonetProperty(EchonetPropertyCode.serial_number)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_serial_number(res.edt)

    def get_current_time_setting(self) -> datetime.time:
        req = EchonetProperty(EchonetPropertyCode.current_time_setting)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_current_time_setting(res.edt)

    def get_current_date_setting(self) -> datetime.date:
        req = EchonetProperty(EchonetPropertyCode.current_date_setting)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_current_date_setting(res.edt)

    def get_properties_for_status_notification(self) -> set[EchonetPropertyCode | int]:
        req = EchonetProperty(EchonetPropertyCode.properties_for_status_notification)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_property_map(res.edt)

    def get_properties_to_set_values(self) -> set[EchonetPropertyCode | int]:
        req = EchonetProperty(EchonetPropertyCode.properties_to_set_values)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_property_map(res.edt)

    def get_properties_to_get_values(self) -> set[EchonetPropertyCode | int]:
        req = EchonetProperty(EchonetPropertyCode.properties_to_get_values)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_property_map(res.edt)

    def get_route_b_id(self) -> dict[str, bytes]:
        req = EchonetProperty(EchonetPropertyCode.route_b_id)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_route_b_id(res.edt)

    def get_one_minute_measured_cumulative_energy(self) -> dict[str, datetime.datetime |
                                                                     dict[str, int | float | None]]:
        req = EchonetProperty(EchonetPropertyCode.one_minute_measured_cumulative_energy)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_one_minute_measured_cumulative_energy(res.edt,
                                                                             self.energy_unit,
                                                                             self.energy_coefficient)

    def get_coefficient_for_cumulative_energy(self) -> int:
        req = EchonetProperty(EchonetPropertyCode.coefficient_for_cumulative_energy)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_coefficient_for_cumulative_energy(res.edt)

    def get_number_of_effective_digits_for_cumulative_energy(self) -> int:
        req = EchonetProperty(EchonetPropertyCode.number_of_effective_digits_for_cumulative_energy)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_number_of_effective_digits_for_cumulative_energy(res.edt)

    def get_measured_cumulative_energy(self,
                                       reverse: bool = False,
                                       ) -> int | float:
        if reverse is False:
            epc = EchonetPropertyCode.measured_cumulative_energy
        else:
            epc = EchonetPropertyCode.measured_cumulative_energy_reversed

        req = EchonetProperty(epc)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_measured_cumulative_energy(res.edt,
                                                                  self.energy_unit,
                                                                  self.energy_coefficient)

    def get_unit_for_cumulative_energy(self) -> int | float:
        req = EchonetProperty(EchonetPropertyCode.unit_for_cumulative_energy)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_unit_for_cumulative_energy(res.edt)

    def get_historical_cumulative_energy_1(self,
                                           day: int = 0,
                                           reverse: bool = False,
                                           ) -> list[dict[str, datetime.datetime | dict[str, int | float | None]]]:
        self.set_day_for_historical_data_1(day)

        if reverse is False:
            epc = EchonetPropertyCode.historical_cumulative_energy_1
        else:
            epc = EchonetPropertyCode.historical_cumulative_energy_1_reversed

        req = EchonetProperty(epc)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_historical_cumulative_energy_1(res.edt,
                                                                      self.energy_unit,
                                                                      self.energy_coefficient)

    def set_day_for_historical_data_1(self,
                                      day: int = 0,
                                      ) -> None:
        edt = EchonetDataBuilder.build_edata_to_set_day_for_historical_data_1(day)
        req = EchonetPropertyWithData(EchonetPropertyCode.day_for_historical_data_1, edt)
        self.__request_to_set([req])

    def get_day_for_historical_data_1(self) -> int:
        req = EchonetProperty(EchonetPropertyCode.day_for_historical_data_1)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_day_for_historical_data_1(res.edt)

    def get_instantaneous_power(self) -> float:
        req = EchonetProperty(EchonetPropertyCode.instantaneous_power)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_instantaneous_power(res.edt)

    def get_instantaneous_current(self) -> dict[str, float]:
        req = EchonetProperty(EchonetPropertyCode.instantaneous_current)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_instantaneous_current(res.edt)

    def get_cumulative_energy_measured_at_fixed_time(self,
                                                     reverse: bool = False,
                                                     ) -> dict[str, datetime.datetime | int | float]:
        if reverse is False:
            epc = EchonetPropertyCode.cumulative_energy_measured_at_fixed_time
        else:
            epc = EchonetPropertyCode.cumulative_energy_measured_at_fixed_time_reversed

        req = EchonetProperty(epc)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_cumulative_energy_measured_at_fixed_time(res.edt,
                                                                                self.energy_unit,
                                                                                self.energy_coefficient)

    def get_historical_cumulative_energy_2(self,
                                           timestamp: datetime.datetime = None,
                                           num_of_data_points: int = 12,
                                           ) -> list[dict[str, datetime.datetime |
                                                               dict[str, int | float | None]]]:
        if timestamp is None:
            timestamp = datetime.datetime.now()

        self.set_time_for_historical_data_2(timestamp, num_of_data_points)

        req = EchonetProperty(EchonetPropertyCode.historical_cumulative_energy_2)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_historical_cumulative_energy_2(res.edt,
                                                                      self.energy_unit,
                                                                      self.energy_coefficient)

    def set_time_for_historical_data_2(self,
                                       timestamp: datetime.datetime,
                                       num_of_data_points: int = 12,
                                       ) -> None:
        edt = EchonetDataBuilder.build_edata_to_set_time_for_historical_data_2(timestamp,
                                                                               num_of_data_points)
        req = EchonetPropertyWithData(EchonetPropertyCode.time_for_historical_data_2, edt)
        self.__request_to_set([req])

    def get_time_for_historical_data_2(self) -> dict[str: datetime.datetime | None,
                                                     str: int]:
        req = EchonetProperty(EchonetPropertyCode.time_for_historical_data_2)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_time_for_historical_data_2(res.edt)

    def get_historical_cumulative_energy_3(self,
                                           timestamp: datetime.datetime = None,
                                           num_of_data_points: int = 10,
                                           ) -> list[dict[str, datetime.datetime |
                                                               dict[str, int | float | None]]]:
        if timestamp is None:
            timestamp = datetime.datetime.now()

        self.set_time_for_historical_data_3(timestamp, num_of_data_points)

        req = EchonetProperty(EchonetPropertyCode.historical_cumulative_energy_3)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_historical_cumulative_energy_3(res.edt,
                                                                      self.energy_unit,
                                                                      self.energy_coefficient)

    def set_time_for_historical_data_3(self,
                                       timestamp: datetime.datetime,
                                       num_of_data_points: int = 10,
                                       ) -> None:
        edt = EchonetDataBuilder.build_edata_to_set_time_for_historical_data_3(timestamp,
                                                                               num_of_data_points)
        req = EchonetPropertyWithData(EchonetPropertyCode.time_for_historical_data_3, edt)
        self.__request_to_set([req])

    def get_time_for_historical_data_3(self) -> dict[str, datetime.datetime | None | int]:
        req = EchonetProperty(EchonetPropertyCode.time_for_historical_data_3)
        res = self.__request_to_get([req])[0]
        return EchonetDataParser.parse_time_for_historical_data_3(res.edt)

    class DayForHistoricalData1(TypedDict, total=False):
        day: int

    class TimeForHistoricalData2(TypedDict, total=False):
        timestamp: datetime.datetime
        num_of_data_points: int

    class TimeForHistoricalData3(TypedDict, total=False):
        timestamp: datetime.datetime
        num_of_data_points: int

    def request_to_set(self,
                       day_for_historical_data_1: DayForHistoricalData1 | None = None,
                       time_for_historical_data_2: TimeForHistoricalData2 | None = None,
                       time_for_historical_data_3: TimeForHistoricalData3 | None = None) -> None:
        properties_with_data = []
        if day_for_historical_data_1 is not None:
            edt = EchonetDataBuilder.build_edata_to_set_day_for_historical_data_1(**day_for_historical_data_1)
            properties_with_data.append(EchonetPropertyWithData(EchonetPropertyCode.day_for_historical_data_1, edt))
        if time_for_historical_data_2 is not None:
            edt = EchonetDataBuilder.build_edata_to_set_time_for_historical_data_2(**time_for_historical_data_2)
            properties_with_data.append(EchonetPropertyWithData(EchonetPropertyCode.time_for_historical_data_2, edt))
        if time_for_historical_data_3 is not None:
            edt = EchonetDataBuilder.build_edata_to_set_time_for_historical_data_3(**time_for_historical_data_3)
            properties_with_data.append(EchonetPropertyWithData(EchonetPropertyCode.time_for_historical_data_3, edt))

        self.__request_to_set(properties_with_data)

    def request_to_get(self,
                       properties: set[EchonetPropertyCode]) -> dict[EchonetPropertyCode, Any]:
        results = self.__request_to_get([EchonetProperty(epc) for epc in properties])
        parsed_results = {}
        for r in results:
            try:
                parser = parser_map[r.epc]
            except KeyError:
                raise MomongaRuntimeError(f"No parser found for EPC: %X" % r.epc)

            sig = inspect.signature(parser)
            args = sig.parameters.keys()
            if "energy_unit" in args and "energy_coefficient" in args:
                parsed_results[r.epc] = parser(r.edt, self.energy_unit, self.energy_coefficient)
            else:
                parsed_results[r.epc] = parser(r.edt)

        return parsed_results



from enum import Enum


class DeviceType(Enum):
    """
    Additional Wi-SUN Modules are included here for completeness but not all are supported.
    Web archive of ROHM Wi-SUN B-Route modules as of November 29, 2025:
    https://web.archive.org/web/20251129002611/https://www.chip1stop.com/sp/products/rohm_wi-sun-module
    """
    BP35A1 = 1  # Internal IC for RL7023 Stick-D/IPS. Now marked Obsolete and EOL
    BP35C0 = 2  # Wi-SUN/HAN Module. Not currently supported in Momonga
    BP35C1 = 3  # Wi-SUN/E-HAN Module. Not currently supported in Momonga
    BP35C2 = 4  # Internal IC for RS-WSUHA-P and RL7023 Stick-D/DSS


class MomongaError(Exception):
    pass


class MomongaSkCommandExecutionFailure(MomongaError):
    pass


class MomongaSkCommandUnknownError(MomongaSkCommandExecutionFailure):
    pass


class MomongaSkCommandUnsupported(MomongaSkCommandExecutionFailure):
    pass


class MomongaSkCommandInvalidArgument(MomongaSkCommandExecutionFailure):
    pass


class MomongaSkCommandInvalidSyntax(MomongaSkCommandExecutionFailure):
    pass


class MomongaSkCommandSerialInputError(MomongaSkCommandExecutionFailure):
    pass


class MomongaSkCommandFailedToExecute(MomongaSkCommandExecutionFailure):
    pass


class MomongaSkScanFailure(MomongaError):
    pass


class MomongaSkJoinFailure(MomongaError):
    pass


class MomongaRuntimeError(RuntimeError):
    pass


class MomongaTimeoutError(TimeoutError):
    pass


class MomongaKeyError(KeyError):
    pass


class MomongaNeedToReopen(MomongaError):
    pass


class MomongaResponseNotExpected(MomongaError):
    pass


class MomongaResponseNotPossible(MomongaError):
    pass



import logging
import threading
import queue
import serial

from typing import Self

from .momonga_exception import (MomongaError,
                                MomongaTimeoutError,
                                MomongaSkCommandUnknownError,
                                MomongaSkCommandUnsupported,
                                MomongaSkCommandInvalidArgument,
                                MomongaSkCommandInvalidSyntax,
                                MomongaSkCommandSerialInputError,
                                MomongaSkCommandFailedToExecute,
                                MomongaSkScanFailure,
                                MomongaSkJoinFailure)
from .momonga_response import (SkVerResponse,
                               SkAppVerResponse,
                               SkInfoResponse,
                               SkScanResponse,
                               SkLl64Response)
from .momonga_device_enum import DeviceType

logger = logging.getLogger(__name__)


class MomongaSkWrapper:
    def __init__(self,
                 dev: str,
                 baudrate: int = 115200,
                 ) -> None:
        self.dev = dev
        self.baudrate = baudrate

        # the following value will be set a pyserial object.
        self.ser = None
        self.publisher_th_breaker = False
        self.publisher_th = None
        self.subscribers = {'cmd_exec_q': queue.Queue()}
        self.device_type: DeviceType = DeviceType.BP35C2

    def __enter__(self) -> Self:
        return self.open()

    def __exit__(self, type, value, traceback) -> None:
        self.close()

    def open(self) -> Self:
        self.ser = serial.Serial(self.dev, self.baudrate)

        try:
            # to drop garbage data in the buffer.
            self.__clear_buf()

            # to check udp payloads returned from the wi-sun module are in ascii format.
            if self.__exec_ropt() != 1:
                logger.warning("Executing 'WOPT 01\\r' command to make the Wi-SUN module return UDP payloads "
                                "in ASCII format. Note: WOPT command can only be executed a limited number of times. "
                                "This configuration is saved in the Wi-SUN module, so this log message should "
                                "no longer appear.")
                self.__exec_wopt(1)  # to make the wi-sun module return udp payloads in ascii format.
        except MomongaSkCommandUnsupported:
            logger.info('ROPT command is unsupported on this hardware. Assuming ASCII output mode.')
        except Exception:
            self.close()
            raise

        for q in self.subscribers.values():
            while not q.empty():
                q.get()

        self.publisher_th_breaker = False  # set True when you want to stop the publisher.
        self.publisher_th = threading.Thread(target=self.received_packet_publisher, daemon=True)
        self.publisher_th.start()

        # Detects device type
        self.detect_device()

    def close(self) -> None:
        if self.publisher_th is not None:
            self.publisher_th_breaker = True
            self.publisher_th.join()
            self.publisher_th = None
        if self.ser is not None and not self.ser.closed:
            self.ser.close()

    def __clear_buf(self) -> None:  # do not call this after open().
        self.ser.write(b'\r\n')
        self.ser.flush()
        timeout = self.ser.timeout
        self.ser.timeout = 2  # will wait the specified seconds.
        while self.ser.read():
            # this loop clears garbage data if it exists.
            pass
        # to undo the timeout.
        self.ser.timeout = timeout

    def __exec_ropt(self) -> int:  # do not call this after open().
        self.ser.write(b'ROPT\r')
        self.ser.flush()
        res = b''
        ok = b'OK '
        fail = b'FAIL'
        while True:
            res += self.ser.read()
            if ok in res and res.endswith(b'\r'):
                break
            elif fail in res and res.endswith(b'\r'):
                decoded = res.decode(errors='replace')
                for line in decoded.splitlines():
                    if line.startswith('FAIL'):
                        self.__raise_fail_response('ROPT', line)
                raise MomongaError('Unexpected ROPT response: %s' % decoded)
        return int(res[res.index(ok) + len(ok):-1].decode())

    def __exec_wopt(self,
                    opt: int,
                    ) -> None:  # do not call this after open().
        supported_opts = (0,  # binary mode
                          1,  # hex ascii mode
                          )
        if opt not in supported_opts:
            raise MomongaError('WOPT command dose not support the given option: %02d' % opt)

        self.ser.write(('WOPT %02d\r' % opt).encode())
        self.ser.flush()
        res = b''
        while True:
            res += self.ser.read()
            if b'OK\r' in res:
                break
        return

    def __readline(self,
                   timeout: int | None = None,
                   ) -> str:
        org_timeout = self.ser.timeout
        self.ser.timeout = timeout
        data_bytes = self.ser.readline()
        self.ser.timeout = org_timeout
        if data_bytes != b'':
            logger.debug('<<< %s' % data_bytes)
        line = data_bytes.decode().split('\r\n')[0]
        return line

    def received_packet_publisher(self) -> None:
        logger.debug('A received packet publisher has been started.')
        while True:
            if self.publisher_th_breaker is True:
                break
            line = self.__readline(timeout=1)
            if line == '':
                continue
            for q in self.subscribers.values():
                q.put(line)  # will dispatch the line to each subscriber
        logger.debug('The received packet publisher has been stopped.')

    def __writeline(self,
                    line: str,
                    payload: bytes | None = None,
                    ) -> None:
        if payload is not None:
            data_bytes = (line + ' ').encode() + payload
        else:
            data_bytes = (line + '\r\n').encode()
        self.ser.write(data_bytes)
        logger.debug('>>> %s' % data_bytes)
        self.ser.flush()

    def exec_command(self,
                     command: list[str],
                     wait_until: str | list[str] = 'OK',
                     timeout: int | None = None,
                     payload: bytes | None = None,
                     ) -> list[str]:
        command = ' '.join([c for c in command if c is not None])

        if type(wait_until) is str:
            wait_until = [wait_until]

        subscriber_q = self.subscribers['cmd_exec_q']
        while not subscriber_q.empty():
            subscriber_q.get()

        self.__writeline(command, payload)

        res = []
        while True:
            r = subscriber_q.get(timeout=timeout)
            if r.startswith('ERXUDP'):
                continue

            if r == '':
                raise MomongaTimeoutError('The command timed out: %s' % (command))
            elif r[:4] == 'FAIL':
                self.__raise_fail_response(command, r)
            else:
                res.append(r)
                matched = False
                for w in wait_until:
                    if r.startswith(w):
                        matched = True
                        break
                if matched is True:
                    break
        return res

    def __raise_fail_response(self, command: str, r: str) -> None:
        error_code = int(r[7:10])
        if 1 <= error_code <= 3:
            raise MomongaSkCommandUnknownError('Unknown error code %s: %s' % (error_code, command))
        elif error_code == 4:
            raise MomongaSkCommandUnsupported('Unsupported command: %s' % (command))
        elif error_code == 5:
            raise MomongaSkCommandInvalidArgument('Invalid argument: %s' % (command))
        elif error_code == 6:
            raise MomongaSkCommandInvalidSyntax('Invalid syntax: %s' % (command))
        elif 7 <= error_code <= 8:
            raise MomongaSkCommandUnknownError('Unknown error code %s: %s' % (error_code, command))
        elif error_code == 9:
            raise MomongaSkCommandSerialInputError('Serial input error: %s' % (command))
        elif error_code == 10:
            raise MomongaSkCommandFailedToExecute(
                'The specified command was accepted but failed to execute: %s' % (command))
        else:
            raise MomongaSkCommandUnknownError('Unknown error code %s: %s' % (error_code, command))

    def skver(self) -> SkVerResponse:
        res = self.exec_command(['SKVER'])
        return SkVerResponse(res)

    def skappver(self) -> SkAppVerResponse:
        res = self.exec_command(['SKAPPVER'])
        return SkAppVerResponse(res)

    def skreset(self) -> None:
        self.exec_command(['SKRESET'])

    def skinfo(self) -> SkInfoResponse:
        res = self.exec_command(['SKINFO'])
        return SkInfoResponse(res)

    def sksreg(self,
               reg: str,
               val: str | int | bytes,
               ) -> None:
        if type(val) is int:
            val = '%X' % val
        elif type(val) is bytes:
            val = val.hex().upper()
        self.exec_command(['SKSREG', reg, val])

    def sksetrbid(self,
                  rbid: str,
                  ) -> None:
        self.exec_command(['SKSETRBID', rbid])

    def sksetpwd(self,
                 pwd: str,
                 ) -> None:
        self.exec_command(['SKSETPWD', '%X' % len(pwd), pwd])

    def skscan(self,
               retry: int = 3,
               ) -> SkScanResponse:
        duration = 6
        for _ in range(retry):
            logger.debug('Trying to scan a PAN... Duration: %d' % duration)
            cmd = []
            match self.device_type:
                case DeviceType.BP35A1:
                    cmd = ['SKSCAN', '2', 'FFFFFFFF', str(duration)]
                case DeviceType.BP35C2:
                    cmd = ['SKSCAN', '2', 'FFFFFFFF', str(duration), '0']
                case _:
                    logger.warning('Unknown device type "%s" detected in skscan. Assuming BP35C2 behavior.', self.device_type)
                    cmd = ['SKSCAN', '2', 'FFFFFFFF', str(duration), '0']
            res = self.exec_command(cmd, 'EVENT 22')
            # estimated execution time: 0.0096s*(2^(DURATION=6)+1)*28 = 17.5s
            # estimated execution time: 0.0096s*(2^(DURATION=7)+1)*28 = 34.7s
            # estimated execution time: 0.0096s*(2^(DURATION=8)+1)*28 = 69.1s
            if 'EPANDESC' in res:
                return SkScanResponse(res, self.device_type)
            duration += 1
        raise MomongaSkScanFailure('Could not find the specified PAN.')

    def skll64(self,
               mac_addr: bytes,
               ) -> SkLl64Response:
        res = self.exec_command(['SKLL64', mac_addr.hex().upper()], 'FE80:')
        return SkLl64Response(res)

    def skjoin(self,
               ip6_addr: str,
               retry: int = 3,
               ) -> None:
        for _ in range(retry):
            logger.debug('Trying to establish a PANA session...')
            res = self.exec_command(['SKJOIN', ip6_addr], ['EVENT 24', 'EVENT 25'])
            # extimated execution time: 2s + 4s + 8s + 8s + 8s + 8s + 8s = 38s ~ 40s
            if res[-1].startswith('EVENT 25'):
                logger.debug('A PANA Session has been established.')
                return
        raise MomongaSkJoinFailure('Could not establish a PANA session.')

    def skterm(self) -> None:
        logger.debug('Trying to terminate the session...')
        res = self.exec_command(['SKTERM'], ['EVENT 27', 'EVENT 28'])
        if res[-1].startswith('EVENT 28'):
            logger.warning('There was no session to terminate.')

    def sksendto(self,
                 ip6_addr: str,
                 data: bytes,
                 handle: int = 1,
                 port: int = 0x0E1A,
                 sec: int = 2,
                 side: int = 0,
                 ) -> None:
        match self.device_type:
            case DeviceType.BP35A1:
                self.exec_command(['SKSENDTO', str(handle), ip6_addr, '%04X' % port,
                                   str(sec), '%04X' % len(data)],
                                  payload=data)
            case DeviceType.BP35C2:
                self.exec_command(['SKSENDTO', str(handle), ip6_addr, '%04X' % port,
                                   str(sec), str(side), '%04X' % len(data)],
                                  payload=data)
            case _:
                logger.warning('Unknown device type "%s" detected in sksendto. Assuming BP35C2 behavior.', self.device_type)
                self.exec_command(['SKSENDTO', str(handle), ip6_addr, '%04X' % port,
                                   str(sec), str(side), '%04X' % len(data)],
                                  payload=data)

    def detect_device(self):
        logger.debug('Trying to detect device...')
        dev_info = self.skinfo()
        if dev_info.side == 65534:
            logger.debug('Device type is BP35A1.')
            self.device_type = DeviceType.BP35A1
        elif dev_info.side < 2:
            logger.debug('Device type is BP35C2.')
            self.device_type = DeviceType.BP35C2
        else:
            logger.warning('Device type is UNKNOWN. Assuming BP35C2.')
            self.device_type = DeviceType.BP35C2



from .momonga import (
    Momonga,
    EchonetPropertyCode,
    EchonetProperty,
    EchonetPropertyWithData,
    logger,
    session_manager_logger,
    sk_wrapper_logger,
)

from .momonga_exception import *



import setuptools


with open('README.md', 'r') as f:
    long_description = f.read()


setuptools.setup(
    name='momonga',
    version='0.3.0',
    description='Python Route B Library: A Communicator for Low-voltage Smart Electric Energy Meters',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/nbtk/momonga',
    author='nbtk',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
    ],
    packages=setuptools.find_packages(),
    install_requires=[
        'pyserial~=3.5',
    ],
    python_requires='>=3.11',
)



import logging

from .momonga_exception import MomongaKeyError
from .momonga_device_enum import DeviceType

logger = logging.getLogger(__name__)


class MomongaSkResponseBase:
    def __init__(self, res):
        self.raw_response = res
        self.decode()

    def decode(self):
        pass

    def extract(self, key):
        for elm in reversed(self.raw_response):
            if key in elm:
                return elm
        raise MomongaKeyError(key)


class MomongaSkDeviceDependentResponse(MomongaSkResponseBase):
    def __init__(self, res, device_type: DeviceType):
        self.device_type = device_type
        super().__init__(res)


class SkVerResponse(MomongaSkResponseBase):
    def decode(self):
        res_list = self.extract('EVER').split()
        self.stack_ver = res_list[1]


class SkAppVerResponse(MomongaSkResponseBase):
    def decode(self):
        res_list = self.extract('EAPPVER').split()
        self.app_ver = res_list[1]


class SkInfoResponse(MomongaSkResponseBase):
    def decode(self):
        res_list = self.extract('EINFO').split()
        self.ip6_addr = res_list[1]
        self.mac_addr = bytes.fromhex(res_list[2])
        self.channel = int(res_list[3], 16)
        self.pan_id = bytes.fromhex(res_list[4])
        self.side = int(res_list[5], 16)


class SkScanResponse(MomongaSkDeviceDependentResponse):
    def decode(self):
        self.channel = int(self.extract('Channel:').split(':')[-1], 16)
        self.channel_page = int(self.extract('Channel Page:').split(':')[-1], 16)
        self.pan_id = bytes.fromhex(self.extract('Pan ID:').split(':')[-1])
        self.mac_addr = bytes.fromhex(self.extract('Addr:').split(':')[-1])
        self.lqi = int(self.extract('LQI:').split(':')[-1], 16)
        self.rssi = 0.275 * self.lqi - 104.27
        match self.device_type:
            case DeviceType.BP35A1:
                self.side = None
            case DeviceType.BP35C2:
                self.side = int(self.extract('Side:').split(':')[-1], 16)
            case _:
                logger.warning('Unknown device type "%s" detected in SkScanResponse. Assuming BP35C2 behavior.', self.device_type)
                self.side = int(self.extract('Side:').split(':')[-1], 16)
        self.pair_id = bytes.fromhex(self.extract('PairID:').split(':')[-1])


class SkLl64Response(MomongaSkResponseBase):
    def decode(self):
        self.ip6_addr = self.extract('FE80:')


class SkSendToResponse(MomongaSkDeviceDependentResponse):
    def decode(self):
        self.res_list = self.extract('EVENT 21').split()
        self.event_num = int(self.res_list[1], 16)
        self.src_addr = self.res_list[2]
        match self.device_type:
            case DeviceType.BP35A1:
                self.side = None
                self.param = int(self.res_list[3], 16)
            case DeviceType.BP35C2:
                self.side = int(self.res_list[3], 16)
                self.param = int(self.res_list[4], 16)
            case _:
                logger.warning('Unknown device type "%s" detected in SkSendToResponse. Assuming BP35C2 behavior.', self.device_type)
                self.side = int(self.res_list[3], 16)
                self.param = int(self.res_list[4], 16)


class SkEventRxUdp(MomongaSkDeviceDependentResponse):
    def decode(self):
        self.res_list = self.extract('ERXUDP').split()
        self.src_addr = self.res_list[1]
        self.des_addr = self.res_list[2]
        self.src_port = int(self.res_list[3], 16)
        self.dst_port = int(self.res_list[4], 16)
        self.src_mac = bytes.fromhex(self.res_list[5])
        match self.device_type:
            case DeviceType.BP35A1:
                self.lqi = None
                self.rssi = None
                self.sec = int(self.res_list[6], 16)
                self.side = None
                self.data_len = int(self.res_list[7], 16)
                self.data = bytes.fromhex(self.res_list[8])
            case DeviceType.BP35C2:
                self.lqi = int(self.res_list[6], 16)
                self.rssi = 0.275 * self.lqi - 104.27
                self.sec = int(self.res_list[7], 16)
                self.side = int(self.res_list[8], 16)
                self.data_len = int(self.res_list[9], 16)
                self.data = bytes.fromhex(self.res_list[10])
            case _:
                logger.warning('Unknown device type "%s" detected in SkEventRxUdp. Assuming BP35C2 behavior.', self.device_type)
                self.lqi = int(self.res_list[6], 16)
                self.rssi = 0.275 * self.lqi - 104.27
                self.sec = int(self.res_list[7], 16)
                self.side = int(self.res_list[8], 16)
                self.data_len = int(self.res_list[9], 16)
                self.data = bytes.fromhex(self.res_list[10])
